#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEBUG	  1

#define MAX_EN	256
#define MAX_NM	 32
#define NOT_FD	 -1
#define ERROR	 -1
#define OK	  1
#define NIL	'\0'
 
/* Global variable and class definitions and declarations */

typedef enum {
	HEADERR, TEXTR, MODIFICATIONR, ENDR, DEFINER, REFERENCER
} RECORD_T;

int assignProgAddr();
int str2int(char *str, int start, int end);
char *str2str(char *str, int start, int end);

class SymEntry {
public:
	char *name;
	long value;
	long length;

	// Methods
	
	SymEntry();
	SymEntry(char *str);
};


class ESTAB {
public:
	class SymEntry entries[MAX_EN];
	int   index;

	// Methods
	
	ESTAB();
	long SearchSymbol(char *sym);
	void EnterSymbol(char *sym, long value, long length=0);
};


class Executable {
public:
	FILE	*outf;

	// Methods
	
	Executable();
	~Executable();
	void ProduceCode(char **table, int startAddr, int flag);
};


class mainClass {
public:
	FILE		 *inf;
	int		 progAddr;
	int		 CSAddr;
	int		 CSLength;
	char		 *line;
	char		 **table;
	RECORD_T	 recType;
	class ESTAB	 symTab;
	class Executable ex;

	// Methods
	
	mainClass(char *name);
	void PassI();
	void PassII();
	int ReadLine();
	void ProcessLine();
	void insertIntoTable(int start, int end, int addr, int flag);
};

/* ****************************************************************************
 * FUNCTION DEFINITIONS
 */

/* ****************************************************************************
 * Class SymEntry
 */

SymEntry::SymEntry()
{
	name = (char *) malloc(MAX_NM+1);
};

SymEntry::SymEntry(char *str)
{
	name = (char *) malloc(MAX_NM+1);
	if (strlen(str) <= MAX_NM) {
		strcpy(name, str);
	}
}

/* ****************************************************************************
 * Class ESTAB
 */

ESTAB::ESTAB()
{
	index = 0;
}

long ESTAB::SearchSymbol(char *sym)
{
	long cslength = 0, oldlength = 0, csaddr = 0;

	for (int i = 0; i < index ; ++i) {
		if (!strcmp(sym, entries[i].name))
			return (long) entries[i].value;
		else if (entries[i].length != 0) {
			oldlength = cslength;
			cslength += entries[i].length;
			csaddr = entries[i].value;
		}
	} // Probably there is no need to these operations
	  // They should already be on the table with correct values
	 
	return (long) NOT_FD;
}

void ESTAB::EnterSymbol(char *sym, long value, long length = 0)
{
	if (index >= MAX_EN) {
		fprintf(stderr, "Symbol table is full...\n");
		exit(1);
	}

	strcpy(entries[index].name, sym);
	entries[index].value = value;
	entries[index++].length = length;
	return;
}

/* ****************************************************************************
 * Class Executable
 */

Executable::Executable()
{
	outf = fopen("a_out", "w+");
}


Executable::~Executable()
{
	fclose(outf);
}

void Executable::ProduceCode(char **table, int startAddr, int flag)
{
	int x = 0, y = 0;
	static int cnt = 0;
	static int ind = 0;

	if (cnt == 0) {
//		fprintf(outf, "Memory               Contents\n");
		fprintf(outf, "0x%04x  ", startAddr);
	}

	cout << "OZGUR SANLI ==> " << cnt << " INDEX "  << ind;


	while  (y < 32) {
		if (table[x][y] == NIL)
			break;
		fprintf(outf, "%c", table[x][y]);
		y++; ind++;
		if (y == 32) {
			//cnt++;
			//fprintf(outf, "\n%04x ", startAddr + cnt*16);
			y = 0; x++;
		}
		if (ind == 32) {
			cnt++;
			fprintf(outf, "\n0x%04x ", startAddr + cnt*16);
			ind = 0;

		}
		if ((ind % 8) == 0) fprintf(outf, " ");
	}

	cout << "OZGUR SANLI ==> " << cnt << " INDEX "  << ind;

#ifdef DEBUG
	fprintf(stdout, "Output file is created...\n");
#endif

	return;
}	


/* ****************************************************************************
 * Class mainClass
 */

mainClass::mainClass(char *name)
{
	inf = fopen(name, "r+");
	if (inf == NULL) {
		fprintf(stderr, "Cannot open the input file ...\n");
		exit(1);
	}
	progAddr = 0;
	CSLength = 0;
	CSAddr = 0;
	line = (char *) malloc(MAX_EN+1);
	recType = HEADERR;
	table = (char **) malloc(sizeof(char *)*1024);
	for (int i = 0; i < MAX_EN; ++i)
		table[i] = (char *) malloc(sizeof(char)*32);
}

int mainClass::ReadLine()
{
	int NEXT;

	char str[MAX_EN+1];

NEXT:
	if (fgets(str, MAX_EN, inf) == NULL)
		return ERROR;
	int i = strlen(str);
	str[i-1] = NIL;
	strcpy(line, str);
	if (line[0] == '\n' || line[0] == '.') goto NEXT;
	return OK;
	// Should Return Some integer value 
}

void mainClass::ProcessLine()
{
	if (line[0] == 'H') {
		recType = HEADERR;
		sscanf(line+7, "%x", &CSLength);
		// Is it SEVEN, 7
	} else if (line[0] == 'D')
		recType = DEFINER;
	else if (line[0] == 'R')
		recType = REFERENCER;
	else if (line[0] == 'T')
		recType = TEXTR;
	else if (line[0] == 'M')
		recType = MODIFICATIONR;
	else if (line[0] == 'E')
		recType = ENDR;

	return;
}


void mainClass::PassI()
{
	char *s;

	progAddr = assignProgAddr();
	CSAddr = progAddr;
	while (ReadLine() != ERROR) {
		ProcessLine();
		CSLength = str2int(line, 13,18);
		if (symTab.SearchSymbol(s = str2str(line, 1,6)) == NOT_FD) {
			symTab.EnterSymbol(s, CSAddr, CSLength);
		} else {
			fprintf(stderr, "Redeclaration of symbol %s", s);
			exit(1);
		}
		while (recType != ENDR) {
			ReadLine();
			ProcessLine();
			if (recType == DEFINER) {
				int p = 0;

				while (line[p*12+1] != '\n' && line[p*12+1] != NIL) {
					char *s = str2str(line, p*12+1, p*12+6);
					if (symTab.SearchSymbol(s) == NOT_FD) {
						symTab.EnterSymbol(s, str2int(line, p*12+7, p*12+12)+CSAddr, 0);
					} else {
						fprintf(stderr, "Redeclaration of %s...\n", s);
						exit(1);
					}
					p++;
				}
			}
		}
		CSAddr += CSLength;
	}
	return;
}

void mainClass::PassII()
{
	int x, y, z, a, b, c;
	rewind(inf);
	CSAddr = progAddr;
	static int addr = progAddr;
	static int old = 0;
	static int oldy = 0;
	static int oldc = 0;

	while (ReadLine() != ERROR) {
		ProcessLine();
		x = str2int(line, 7, 12);
		CSLength = str2int(line, 13, 18);
		while (recType != ENDR) {
			ReadLine();
			ProcessLine();
			if (recType == TEXTR) {
				insertIntoTable(str2int(line, 1, 6)+CSAddr, str2int(line, 7, 8), addr, 1);
				addr = str2int(line, 1, 6) + str2int(line, 7, 8) + CSAddr;
			} else if (recType == MODIFICATIONR) {
				char *s;
				int k =  symTab.SearchSymbol(s = str2str(line, 10, 15));
				if (k != NOT_FD) {
					y = str2int(line, 1, 6);
					b = (int) ((y - x) / 16);
					a = (y - x) % 16;
					a *= 2;
					c = str2int(line, 7, 8); // # of half bytes to be modified
					if ((c % 2) != 0)  a++;

					char ozgur[10];
					int tmpa = a, tmpb = b;
					for (int e = 0; e < c; ++e) {
						if (tmpa == 32) {
							tmpb++;
							tmpa = 0;
						}
						ozgur[e] = table[tmpb][tmpa++];
					}
					ozgur[c] = NIL;
					long aa;
					sscanf(ozgur, "%x", &aa);
//					cout << "Before ... " << "aa: " << aa << " k: " << k << "\n";
					if ((old != 0) && (oldy == y) && (y != 0)) {
						if (line[9] == '+')
							aa = old + k;
						else
							aa = old - k;
					} else {
						if (line[9] == '+')
							aa += k;
						else
							aa -= k;
					}
//					cout << "After ==> " << aa << "\n";

					if ((oldy != y) && (oldy != 0)) {
						int m,n;
						m = oldy % 16;
						n = oldy / 16;
						m *= 2;
						if ((oldc % 2) != 0) m++;
						sprintf(ozgur, "%08X", old);

						for (int e = 8 - oldc; e < 8; ++e) {
							if (m == 32) {
								n++;
								m = 0;
							}
							table[n][m++] = ozgur[e];
						}
						old = 0;
						oldy = 0;
					}

					if ((aa >= 0) && (aa < 0x01000000)) { 

						sprintf(ozgur, "%08X", aa);
//						cout << "WRITING " << ozgur << "\n";
						for (int e = 8 - c; e < 8; ++e) {
							if (a == 32) {
								b++;
								a = 0;
							}
							table[b][a++] = ozgur[e];
						}
						old = 0;	
						oldy = 0;
					} else {
						old = aa;
						oldy = y;
						oldc = c;
//						fprintf(stdout, "OLD %X\n", aa);
					}
				} else {
					fprintf(stderr, "Undeclared symbol...\n");
					exit(1);
				}
			}
		}

		if (old != 0) {
			int m, n;
			char ozgur[10];
			m = oldy % 16;
			n = oldy / 16;
			m *= 2;
			if ((oldc % 2) != 0) m++;
			sprintf(ozgur, "%08X", old);

			for (int e = 8 - oldc; e < 8; ++e) {
				if (m == 32) {
					n++;
					m = 0;
				}
				table[n][m++] = ozgur[e];
			}
			old = 0;
			oldy = 0;
		}

		ex.ProduceCode(this->table, progAddr, 1);
		insertIntoTable(0, 0, 0, 0);
		CSAddr += CSLength;
	}
	return;
}

void mainClass::insertIntoTable(int start, int end, int addr, int flag)
{
	static int indx = 0;
	static int indy = 0;

	if (flag == 0) {
		indx = 0;
		indy = 0;
		table[indx][indy] = NIL;
		return;
	}

	for (int i = addr; i < start; ++i) {
		if (indy == 32) {
			indx++;
			indy = 0;
		}
		table[indx][indy] = '.';
		table[indx][indy+1] = '.';
		indy += 2;
	}

	for (int i = 9; i < 9+2*end; ++i) {
		if (indy == 32) {
			indx++;
			indy = 0;
		}
		table[indx][indy] = line[i];
		indy++;
	}

	table[indx][indy] = NIL;
	return;
}

/* ****************************************************************************
 * Global Function Definitions
 */

int assignProgAddr()
{
	return (4*16*256);
}

int str2int(char *str, int start, int end)
{
	char *s = new char(128);
	int i, j;
	for (i = start, j = 0; i <= end; ++i ,++j)
		s[j] = str[i];
	s[j] = '\0';

	int sum;
	sscanf(s, "%x", &sum);

	delete[] s;
	return sum;
}


char *str2str(char *str, int start, int end)
{
	char *s = new char(128);
	int i, j;
	for (i = start, j = 0; i <= end; ++i ,++j)
		s[j] = str[i];
	s[j] = '\0';
	return s;
}

/* **************************************************************************
 * MAIN FUNCTION
 */

int main(int argc, char **argv)
{
	if (argc != 2) {
		fprintf(stderr, "Usage:\n\t%s <input_object_file>\n", argv[0]);
		exit(1);
	}

	class mainClass loadLink(argv[1]);

	loadLink.PassI();
	loadLink.PassII();

	cout << "Executable is produced...\n";
	cout << sizeof(long);
	return 0;
}
